from django.shortcuts import render
from django.http import HttpResponse
from django.utils.safestring import mark_safe
import datetime
from .models import *
import simplejson
def main(request):
    if request.method=="GET":
        models=request.session.get("models",None)
        if not models:
            models = [s['table'] for s in History.objects.values("table").distinct()]
            request.session["models"]=models
        res={"models": models}
        return render(request,"main.html",res)
    if request.method=="POST":
        id = request.POST["id"]
        table = request.POST["table"]
        models = request.session.get("models", None)
        res=fetchChanges(id,table)
        res["models"]=models
        return render(request,"main.html",res)
def get(lst,index,default):
    if index<len(lst): return lst[index]
    return default
def findChanges(old_state,new_state):
    res="<ul>"
    if type(old_state)==type({}):
        for key in old_state:
            if key == "_type":
                if old_state[key]=="datetime":
                    if datetime.datetime.strptime(old_state["value"],"%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%SZ") != new_state:
                        res += "<li> %s ----> %s</li>" % (old_state.get("value"), new_state)
                elif old_state[key]=="date":
                    if datetime.datetime.strptime(old_state["value"],"%Y-%m-%d").strftime("%Y-%m-%d") != new_state:
                        res += "<li> %s ----> %s</li>" % (old_state.get("value"), new_state)
                break
            elif new_state is None:
                res += "<li> %s ----> NONE</li>" % (old_state.get(key))
            elif old_state[key]!=new_state.get(key,None):
                if type(old_state[key]) in [type({}),type([])]:
                    try:
                        c = findChanges(old_state.get(key, {}), new_state.get(key, {}))
                        if c!="<ul></ul>":
                            res+= "<li>"+c+"</li>"
                    except:
                        pass
                else:
                    res+="<li>%s:: %s ----> %s</li>"%(key,old_state.get(key,None),new_state.get(key,None))
    elif type(old_state)==type([]):
        for key in range(len(old_state)):
            if old_state[key] != get(new_state,key,None):
                if type(old_state[key]) in [type({}), type([])]:
                    try:
                        res += findChanges(old_state[key], new_state[key])
                    except:
                        pass
                else:
                    res += "<li>%s:: %s ----> %s</li>" % (key, old_state[key], get(new_state,key,None))
    return res+"</ul>"
def fetchChanges(id,table):
    changes = History.objects.filter(primary_key=id, table=table).order_by("-id")
    rows = []
    for change in changes:
        row = {}
        row["event_time"] = change.done_on
        row["by"] = change.done_by
        row["changes"] = []
        row["name"]=change.name
        row["id"]=change.id
        for key in change.new_state.keys():
            if type(change.new_state[key]) ==type({}) and change.new_state[key].get("_type",None)!=None:

                if change.new_state[key]["_type"]=="datetime":
                    try:
                        change.new_state[key]=datetime.datetime.strptime(change.new_state[key]["value"],"%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%SZ")
                    except ValueError as exp:
                        change.new_state[key]=datetime.datetime.strptime(change.new_state[key]["value"],"%Y-%m-%d").strftime("%Y-%m-%d")
                elif change.new_state[key]["_type"]=="date":
                    change.new_state[key]=datetime.datetime.strptime(change.new_state[key]["value"],"%Y-%m-%d").date().strftime("%Y-%m-%d")
            if change.old_state.get(key, None) != change.new_state.get(key, None):
                if type(change.old_state.get(key, None)) in [type({}), type([])]:
                    keyChanges = findChanges(change.old_state[key], change.new_state[key])
                    if keyChanges!="<ul></ul>":
                        text = "%s: <br/>" % key
                        text += keyChanges
                        row["changes"].append(mark_safe(text))
                else:
                    row["changes"].append(
                        "%s: %s ----> %s" % (key, change.old_state.get(key, None), change.new_state[key]))
        rows.append(row)
    count = len(rows)
    res = {"count": count, "changes": rows, "id": id, "selected_model": table}
    return res

def getChanges(request):
    id = request.GET["id"]
    table = request.GET["table"]
    res=fetchChanges(id,table)
    return HttpResponse(simplejson.dumps(res))

def showChanges(request):
    table = request.POST["table"]
    primary_key = request.POST["id"]
    res = fetchChanges(primary_key, table)
    return render(request,"changes.html", res)




from django.core.management.base import BaseCommand, CommandError
import __future__
from ModelTracker.models import History
import datetime
from django.apps import apps
def getModel(table_name):
    return next((m for m in apps.get_models() if m._meta.db_table==table_name), None)

class Command(BaseCommand):
    help = 'Restore Object to old status'

    def add_arguments(self, parser):
        parser.add_argument('--id', nargs='?', type=str,default=None)
        parser.add_argument("--state",type=str,nargs='?',default="new")
    def handle(self, *args, **options):
        if not options.get("id",None):
            print ("Change ID is needed")
            exit(1)
        print (options)
        h = History.objects.get(id=int(options["id"]))
        model = getModel(h.table)
        if model == None:
            print("Can't find the Model")
            exit(2)
        d=[f.name for f in model._meta.get_fields()]
        if options["state"]=="old": state=h.old_state
        else: state=h.new_state
        keys2del=[]
        for key in state:
            if (key.startswith("_") and "_cache" in key) or (key not in d and not ("_id" in key and key[:-3] in d)):
                keys2del.append(key)
            if type(state[key])==type({}):
                if state[key].get("_type",None) == "datetime":
                    state[key]=datetime.datetime.strptime(state[key]["value"],"%Y-%m-%d %H:%M:%S")
                elif state[key].get("_type",None) == "date":
                    state[key]=datetime.datetime.strptime(state[key]["value"],"%Y-%m-%d")
        for key in keys2del:
            del state[key]

        print(state)
        m=model(**state)
        m.save("CLI",event_name="Restore Record to %s (%s)"%(options["id"],options["state"]))







# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('ModelTracker', '0002_auto_20160929_0517'),
    ]

    # state_operations = [
    #     migrations.AddField(
    #         model_name='history',
    #         name='related_objects',
    #         field=jsonfield.fields.JSONField(null=True),
    #     ),
    #     ]
    #
    # db_operations= [migrations.RunSQL("""ALTER TABLE `ModelTracker_history` ADD COLUMN `related_objects` JSON;""")]

    operations = [
#            migrations.SeparateDatabaseAndState(state_operations=state_operations,database_operations=db_operations)
        ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='History',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, db_column='id')),
                ('table', models.CharField(max_length=255)),
                ('primary_key', models.CharField(max_length=255)),
                ('old_state', jsonfield.fields.JSONField(default={})),
                ('new_state', jsonfield.fields.JSONField(default={})),
                ('done_by', models.CharField(max_length=255)),
                ('done_on', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ModelTracker', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='history',
            name='name',
            field=models.CharField(default='', max_length=255),
        ),
        migrations.AlterField(
            model_name='history',
            name='id',
            field=models.AutoField(serialize=False, primary_key=True),
        ),
    ]



from django.db import models
import sys
from .models import *
import copy
from django.utils import timezone
from django.contrib.admin.utils import NestedObjects
from django.contrib.admin import ModelAdmin
import threading
import datetime

class ModelTracker(models.Model):
    thread = threading.local()
    def __init__(self,*args,**kwargs):
        models.Model.__init__(self, *args, **kwargs)
        self.old_state = copy.deepcopy(self.__dict__)



    def save(self, username='', event_name="",force_insert=False, force_update=False, using=None, update_fields=None):
        types=[]
        if username==None:
            models.Model.save(self,force_insert=force_insert,force_update=force_update,using=using,update_fields=update_fields)
            return
        from inspect import currentframe, getframeinfo
        frameinfo = getframeinfo(currentframe().f_back)
        if "models.py" in frameinfo.filename:
            frameinfo = getframeinfo(currentframe().f_back.f_back)

        if username=='':  username= self.thread.request.user.username
        if sys.version_info > (3,):
            lng=int
        else:
            lng=long
        types=[type("a"),type(1),type({}),type([]),type(("1",2)),type(True),type(lng(1)),type(u"a"),type(1.1),type(None)]
        history = History()
        history.table = self._meta.db_table
        history.done_on = timezone.now()
        history.done_by = username
        history.name=event_name
        x=self.__dict__.copy()
        history.new_state = copy.deepcopy(x)
        history.new_state.pop("old_state")

        if self.pk == None:
            history.old_state = {}
        else:
            history.old_state=self.old_state
        keys2del=[]
        for key in history.old_state:
            if key.startswith("_") and "_cache" in key:
                keys2del.append(key)
                continue

            if type(history.old_state[key]) not in types:
                if hasattr(history.old_state[key],"toJSON"):
                    history.old_state[key]=history.old_state[key].toJSON()
                elif hasattr(history.old_state[key],"pk"):
                    history.old_state[key]= history.old_state[key].pk
                elif type(history.old_state[key])==type(datetime.datetime.now()):
                    dt=history.old_state[key]
                    history.old_state[key]={"_type":"datetime","value":"%s-%s-%s %s:%s:%s"%(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second)}
                elif type(history.old_state[key])==type(datetime.datetime.now().date()):
                    d=history.old_state[key]
                    history.old_state[key]={"_type":"date","value":"%s-%s-%s"%(d.year,d.month,d.day)}
                else:
                    keys2del.append(key)
        for key in keys2del:
            del history.old_state[key]
        keys2del=[]
        for key in history.new_state:
            if key.startswith("_") and "_cache" in key:
                keys2del.append(key)
                continue
            if type(history.new_state[key]) not in types:
                if hasattr(history.new_state[key], "toJSON"):
                    history.new_state[key] = history.new_state[key].toJSON()
                elif hasattr(history.new_state[key], "pk"):
                    history.new_state[key] = history.new_state[key].pk
                elif type(history.new_state[key])==type(datetime.datetime.now()):
                    dt = history.new_state[key]
                    history.new_state[key] = {"_type": "datetime", "value": "%s-%s-%s %s:%s:%s" % (
                    dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)}
                elif type(history.new_state[key]) == type(datetime.datetime.now().date()):
                    d = history.new_state[key]
                    history.new_state[key] = {"_type": "date", "value": "%s-%s-%s" % (d.year, d.month, d.day)}
                else:
                    keys2del.append(key)
        for key in keys2del:
            del history.new_state[key]
        models.Model.save(self,force_insert=force_insert,force_update=force_update,using=using,update_fields=update_fields)
        history.primary_key=self.pk
        history.new_state.pop("_state","")
        history.new_state["__called_through"] = "%s %s" % (frameinfo.filename, frameinfo.lineno)
        history.old_state.pop("_state","")
        history.save()


    def delete(self, username='', event_name="", using=None, ):
        def format(obj):
            return "%s:%s" % (obj._meta.db_table, obj.pk)

        types = []
        if username == None:
            models.Model.delete(self, using=using)
            return
        if username == '':  username = self.thread.request.user.username
        if sys.version_info > (3,):
            lng = int
        else:
            lng = long
        types = [type("a"), type(1), type({}), type([]), type(("1", 2)), type(True), type(lng(1)), type(u"a"),
                 type(1.1), type(None)]
        history = History()
        history.table = self._meta.db_table
        history.done_on = timezone.now()
        history.done_by = username
        if event_name == '': event_name = "Delete"
        history.name = event_name

        history.old_state = self.old_state
        history.primary_key = self.pk
        history.new_state.pop("_state", "")
        history.old_state.pop("_state", "")
        history.new_state={"related_records":[]}

        collector = NestedObjects('default')
        collector.collect([self])
        x = collector.nested(format)
        if len(x) > 1:
            history.new_state["related_records"]=x[1]
        history.save()
        history.new_state.pop("_state", "")
        history.old_state.pop("_state", "")
        models.Model.delete(self, using=using)

    class Meta:
        abstract =True


class TrackerAdmin(ModelAdmin):
    def save_model(self, request, obj, form, change):
        obj.save(request.user.username, "Editing From admin interface")

    def delete_model(self, request, obj):
        obj.delete(username = request.user.username, event_name = "Deleting From admin interface")


from django.db import models
from jsonfield import JSONField
class History(models.Model):
    id=models.AutoField(primary_key=True)
    name=models.CharField(max_length=255,default="")
    table=models.CharField(max_length=255)
    primary_key=models.CharField(max_length=255)
    old_state=JSONField(default={})
    new_state=JSONField(default={})
    done_by=models.CharField(max_length=255)
    done_on=models.DateTimeField(auto_now_add=True)
    def __unicode__(self):
        return self.id




#!/usr/bin/env python

# from distutils.core import setup
# from setuptools import find_packages
#
# setup(name='DjangoModelTracker',
#       version='0.2',
#       description='Track Django Model Objects over time',
#       author='Mohamed El-Kalioby',
#       author_email='mkalioby@mkalioby.com',
#       url='https://github.com/mkalioby/ModelTracker/',
#       packages=('ModelTracker','ModelTracker.migrations'),
#
#    )

from setuptools import find_packages, setup

setup(
    name='django-model-tracker',
    version='1.5.2',
    description='Track Django Model Objects over time',
    author='Mohamed El-Kalioby',
    author_email = 'mkalioby@mkalioby.com',
    url = 'https://github.com/mkalioby/ModelTracker/',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",

    download_url='https://github.com/mkalioby/ModelTracker/',
    license='MIT',
    packages=find_packages(exclude=('TestApp.*', 'TestApp', 'example')),
    install_requires=[
        'Django>=1.7',
        'jsonfield',
        'simplejson'
      ],
    include_package_data=True,
      zip_safe=False, # because we're including static files
        python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 1.7",
        "Framework :: Django :: 1.8",
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.10",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules",
],

)



from django.conf.urls import url
from . import views

urlpatterns = [
    url('^$',views.main,name="main"),
    url('show/',views.showChanges,name="showModelChanges"),
    url('get/',views.getChanges,name="getModelChanges"),
                       ]
